#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI first-and-second-generation Sagnac variables.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from . import core
from . import intervar


#: First-generation :math:`\alpha_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ALPHA1_ETA = core.LISATDICombination.from_string('1321 -1321')

#: First-generation :math:`\beta_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
BETA1_ETA = ALPHA1_ETA.rotated()

#: First-generation :math:`\gamma_1` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
GAMMA1_ETA = BETA1_ETA.rotated()

#: First-generation :math:`\zeta_1` fully-symmetric Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA1_ETA = core.LISATDICombination({
    'eta_12': [(1, ['D_23'])],
    'eta_23': [(1, ['D_31'])],
    'eta_31': [(1, ['D_12'])],
    'eta_13': [(-1, ['D_32'])],
    'eta_32': [(-1, ['D_21'])],
    'eta_21': [(-1, ['D_13'])],
})

#: First-generation :math:`\alpha_1` Sagnac combination.
#:
#: This combination is the composition of ``ALPHA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ALPHA1 = ALPHA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\beta_1` Sagnac combination.
#:
#: This combination is the composition of ``BETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
BETA1 = BETA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\gamma_1` Sagnac combination.
#:
#: This combination is the composition of ``GAMMA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
GAMMA1 = GAMMA1_ETA @ intervar.ETA_SET

#: First-generation :math:`\zeta_1` fully-symmetric Sagnac combination.
#:
#: This combination is the composition of ``ZETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA1 = ZETA1_ETA @ intervar.ETA_SET

#: Second-generation :math:`\alpha_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ALPHA2_ETA = core.LISATDICombination.from_string('1231321 -1321231')

#: Second-generation :math:`\beta_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
BETA2_ETA = ALPHA2_ETA.rotated()

#: Second-generation :math:`\gamma_2` Sagnac combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
GAMMA2_ETA = BETA2_ETA.rotated()

#: Second-generation :math:`\zeta_{21}` fully-symmetric Sagnac combination of first kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA21_ETA = core.LISATDICombination({
    'eta_12': [(1, ['D_23', 'D_32']), (-1, ['D_13', 'D_21', 'D_32'])],
    'eta_23': [(1, ['D_32', 'D_13']), (-1, ['D_12', 'D_31', 'D_13'])],
    'eta_31': [(1, ['D_23', 'D_12']), (-1, ['D_13', 'D_21', 'D_12'])],
    'eta_13': [(1, ['D_12', 'D_31', 'D_23']), (-1, ['D_32', 'D_23'])],
    'eta_32': [(1, ['D_13', 'D_21', 'D_12']), (-1, ['D_23', 'D_12'])],
    'eta_21': [(1, ['D_12', 'D_31', 'D_13']), (-1, ['D_32', 'D_13'])],
})

#: Second-generation :math:`\zeta_{22}` fully-symmetric Sagnac combination of second kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA22_ETA = ZETA21_ETA.rotated()

#: Second-generation :math:`\zeta_{23}` fully-symmetric Sagnac combination of third kind.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
ZETA23_ETA = ZETA22_ETA.rotated()

#: Second-generation :math:`\alpha_2` Sagnac combination.
#:
#: This combination is the composition of ``ALPHA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ALPHA2 = ALPHA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\beta_2` Sagnac combination.
#:
#: This combination is the composition of ``BETA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
BETA2 = BETA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\gamma_2` Sagnac combination.
#:
#: This combination is the composition of ``GAMMA1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
GAMMA2 = GAMMA2_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{21}` fully-symmetric Sagnac combination of first kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA21 = ZETA21_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{22}` fully-symmetric Sagnac combination of second kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA22 = ZETA22_ETA @ intervar.ETA_SET

#: Second-generation :math:`\zeta_{23}` fully-symmetric Sagnac combination of third kind.
#:
#: This combination is the composition of ``ZETA21_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
ZETA23 = ZETA23_ETA @ intervar.ETA_SET



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
"""PyTDI module."""

from .meta import __version__
from .meta import __author__
from .meta import __email__
from .meta import __copyright__

from .core import TDICombination
from .core import LISATDICombination
from .core import LISAClockCorrection

from .interface import Data



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines first and second-generation quasi-orthogonal combinations (A, E, and T).

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from numpy import sqrt
from .michelson import X1, Y1, Z1, X2, Y2, Z2


#: First-generation :math:`A_1` orthogonal combination.
#:
#: This combination is a linear combinations of first-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
A1 = (Z1 - X1) / sqrt(2)

#: First-generation :math:`E_1` orthogonal combination.
#:
#: This combination is a linear combinations of first-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
E1 = (X1 - 2 * Y1 + Z1) / sqrt(6)

#: First-generation :math:`T_1` orthogonal combination.
#:
#: This combination has a reduced sensitivity to gravitational waves,
#: but retain a high response function to instrumental noises. Therefore,
#: it is often dubbed the *null channel*.
#:
#: This combination is a linear combinations of first-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
T1 = (X1 + Y1 + Z1) / sqrt(3)

#: Second-generation :math:`A_2` orthogonal combination.
#:
#: This combination is a linear combinations of second-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
A2 = (Z2 - X2) / sqrt(2)

#: Second-generation :math:`E_2` orthogonal combination.
#:
#: This combination is a linear combinations of second-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
E2 = (X2 - 2 * Y2 + Z2) / sqrt(6)

#: Second-generation :math:`T_1` orthogonal combination.
#:
#: This combination has a reduced sensitivity to gravitational waves,
#: but retain a high response function to instrumental noises. Therefore,
#: it is often dubbed the *null channel*.
#:
#: This combination is a linear combinations of second-generation Michelson
#: combinations, c.f. :ref:`standard-combinations:michelson combinations`.
#:
#: It is function of the beatnote measurements.
T2 = (X2 + Y2 + Z2) / sqrt(3)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI intermediary variables ξ (xi) and η (eta).

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from . import core


#: :math:`\xi_{12}` spacecraft jitter-reducing intermediary variable.
XI_12 = core.LISATDICombination({
    "isi_12": [(1.0, [])],
    "rfi_12": [(0.5, [])],
    "rfi_21": [(0.5, ['D_12'])],
    "tmi_12": [(-0.5, [])],
    "tmi_21": [(-0.5, ['D_12'])],
})

#: :math:`\xi_{23}` spacecraft jitter-reducing intermediary variable.
XI_23 = XI_12.rotated()

#: :math:`\xi_{31}` spacecraft jitter-reducing intermediary variable.
XI_31 = XI_23.rotated()

#: :math:`\xi_{13}` spacecraft jitter-reducing intermediary variable.
XI_13 = XI_12.reflected(1)

#: :math:`\xi_{21}` spacecraft jitter-reducing intermediary variable.
XI_21 = XI_23.reflected(2)

#: :math:`\xi_{31}` spacecraft jitter-reducing intermediary variable.
XI_32 = XI_31.reflected(3)

#: dict: Dictionary of all 6 :math:`\xi_{ij}` variables.
#:
#: This dictionary is intended to be used for composition of combinations.
XI_SET = {
    "xi_12": XI_12, "xi_23": XI_23, "xi_31": XI_31,
    "xi_13": XI_13, "xi_21": XI_21, "xi_32": XI_32,
}

#: :math:`\eta_{12}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_12_XI = core.LISATDICombination({
    "xi_12": [(1.0, [])],
    "rfi_23": [(-0.5, ['D_12'])],
    "rfi_21": [(0.5, ['D_12'])],
}, allow_reflections=False)

#: :math:`\eta_{23}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_23_XI = ETA_12_XI.rotated()

#: :math:`\eta_{31}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_31_XI = ETA_23_XI.rotated()

#: :math:`\eta_{13}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_13_XI = core.LISATDICombination({
    "xi_13": [(1.0, [])],
    "rfi_12": [(0.5, [])],
    "rfi_13": [(-0.5, [])],
}, allow_reflections=False)

#: :math:`\eta_{21}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_21_XI = ETA_13_XI.rotated()

#: :math:`\eta_{32}` intermediary variable to reduce the problem to 3 lasers.
#:
#: This combination is defined as a function of :math:`\xi` measurements in
#: addition to the reference beatnotes.
ETA_32_XI = ETA_21_XI.rotated()

#: dict: Dictionary of all 6 :math:`\eta_{ij}` variables.
#:
#: The :math:`\eta_{ij}` are defined as a function of the :math:`\xi` measurements in
#: addition to the reference beatnotes.
#:
#: This dictionary is intended to be used for composition of combinations.
ETA_XI_SET = {
    "eta_12": ETA_12_XI, "eta_23": ETA_23_XI, "eta_31": ETA_31_XI,
    "eta_13": ETA_13_XI, "eta_21": ETA_21_XI, "eta_32": ETA_32_XI,
}

#: dict: Dictionary of all 6 :math:`\eta_{ij}` variables.
#:
#: The :math:`\eta_{ij}` are defined as a function of the beatnotes.
#:
#: This dictionary is intended to be used for composition of combinations.
ETA_SET = {
    "eta_12": ETA_12_XI @ XI_SET, "eta_23": ETA_23_XI @ XI_SET, "eta_31": ETA_31_XI @ XI_SET,
    "eta_13": ETA_13_XI @ XI_SET, "eta_21": ETA_21_XI @ XI_SET, "eta_32": ETA_32_XI @ XI_SET,
}



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines the core class `TDICombination`, used to build combinations.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

import copy
import itertools
import logging
from . import dsp
from . import naming

logger = logging.getLogger(__name__)


class TDICombination():
    """Defines a time-delay interferometry (TDI) combination.

    To create a combination, pass a component dictionary containing the measurements
    as keys and time-shift polynomials as values.

    Time-shift polynomials are defined as a list of couples ``(factor, operators)``,
    where ``factor`` is a scalar factor, and ``operators`` is a list of nested
    time-shift operators.

    .. code-block:: python

        mytdi = TDICombination({
            'y_12': [(1, ['D_12', 'D_23']), (-1, ['A_12'])],
            'y_21': [(0.5, ['A_23', 'D_23'])],
        })

    Delay operators are represented by the string ``D_ij``, where :math:`i` and
    :math:`j` are the usual indices, and advancement operators by ``A_ij``.

    Args:
        components (dict): dictionary of components
    """

    def __init__(self, components):
        logger.info("Initializing combination with components '%s'", components)
        #: dict: Dictionary of components.
        #:
        #: This is the dictionary of measurements associated with the list of time-shift
        #: polynomials, as couples ``(factor, operators)``.
        #:
        #: Refer to the documentation of :class:`pytdi.TDICombination` for more information.
        self.components = dict(components)

    @classmethod
    def from_string(cls, string, var_name='eta', half_way_shift=True):
        """Initialize a combination from virtual photon paths.

        Following the geometrical interpretation proposed by M. Vallisneri, you can
        create a combination from the virtual interference of multiple beams.

        .. code-block:: python

            X1 = LISATDICombination.from_string('13121 -13121')

        Args:
            string (str): virtual photon paths
            var_name (str): name of the measurement
            half_way_shift (bool): whether to compensate for half of the shifts

        Returns:
            :obj:`pytdi.TDICombination`: A new combination.
        """
        # Vanishing combination for empty string
        if not string:
            return cls({})

        beams = string.split(' ')
        combination = cls({})
        total_shifts = []

        for beam in beams:
            if beam[0] == '-':
                is_delay = True
                beam = beam[1:]
            else:
                is_delay = False

            indices = [int(i) for i in beam]
            for i, j in zip(indices[:-1], indices[1:]):
                if is_delay:
                    combination -= total_shifts * cls({naming.join_indices(var_name, i, j): [(1, [])]})
                    total_shifts.append(naming.join_indices("D", i, j))
                else:
                    combination += total_shifts * cls({
                        naming.join_indices(var_name, j, i): [(1, [naming.join_indices("A", i, j)])]
                    })
                    total_shifts.append(naming.join_indices("A", i, j))

        if half_way_shift:
            half_shifts_inv = []
            for operator in reversed(total_shifts[:len(total_shifts)//2]):
                name, i, j = naming.split_indices(operator)
                name_inv = 'D' if name == 'A' else 'A'
                half_shifts_inv.append(naming.join_indices(name_inv, j, i))
            return half_shifts_inv * combination
        return combination

    @property
    def measurements(self):
        """Return the set of measurements in the combination."""
        return set(self.components.keys())

    @property
    def delays(self):
        """Return the set of all delay operators in the combination."""
        delays = set()
        for polynomial in self.components.values():
            for _, operators in polynomial:
                for operator in operators:
                    name, _, _ = naming.split_indices(operator)
                    if name == "D":
                        delays.add(operator)
        return delays

    @property
    def advancements(self):
        """Return the set of all advancement operators in the combination."""
        advancements = set()
        for polynomial in self.components.values():
            for _, operators in polynomial:
                for operator in operators:
                    name, _, _ = naming.split_indices(operator)
                    if name == "A":
                        advancements.add(operator)
        return advancements

    def build_shifts(self, delays, fs, delay_derivatives=None, order=5, delta=1e-12, maxiter=10):
        """Precompute total shifts.

        Nested time-shifts are iteratively constructed from atomic time-shifts.
        If `delay_derivatives=None`, no shift derivatives are returned. Otherwise,
        derivatives of all nested time-shifts are calculated and returned.

        .. code-block:: python

            shifts, derivatives = mytdi.build_shifts(
                {'d_12': t12_array, 'd_21': t21_array}, fs,
            )

        Args:
            delays (dict): dictionary of delays [s]
            fs (double): sampling frequency [Hz]
            delay_derivatives: dictionary of delay time derivatives [s/s]
            order (int): Lagrange interpolation orders for delay interpolations
            delta (float): allowed timing accuracy for advancements calculation
            maxiter (int): maximum iterations for advancements calculation

        Returns:
            tuple of :obj:`np.ndarray`: Time-shifts and time-shift derivatives.
        """
        order = int(order)
        logger.debug("Using interpolation order '%d' for time shifts", order)

        shifts = {(): 0}
        derivatives = {(): 0}
        logger.debug("Computing delays")
        for delay in self.delays:
            logger.debug("Computing delay '%s'", delay)
            # split off indices from delay operator D_ij
            _, i, j = naming.split_indices(delay)
            shifts[(delay, )] = -delays[naming.join_indices('d', i, j)]
            if delay_derivatives:
                logger.debug("Computing delay derivative '%s'", delay)
                derivatives[(delay, )] = -delay_derivatives[naming.join_indices('d', i, j)]
        logger.debug("Computing advancements")
        for advancement in self.advancements:
            logger.debug("Computing advancement '%s'", advancement)
            _, i, j = naming.split_indices(advancement)
            delay = naming.join_indices('d', j, i)
            shifts[(advancement, )] = dsp.calculate_advancements(delays[delay], fs, order, delta, maxiter)
            if delay_derivatives:
                logger.debug("Computing advancement derivative '%s'", advancement)
                dotd_advanced = dsp.timeshift(delay_derivatives[delay], shifts[(advancement, )] * fs)
                derivatives[(advancement, )] = dotd_advanced / (1 - dotd_advanced)

        # Find set of all nested operators
        nested_operators = set()
        for _, polynomial in self.components.items():
            for _, operators in polynomial:
                if len(operators) > 1:
                    nested_operators.add(tuple(operators))

        for operator in nested_operators:
            # Compute the global time shift to apply
            nested_shifts = 0
            nested_derivatives = 0
            for atom in operator:
                if delay_derivatives:
                    atomic_derivatives = derivatives[(atom, )]
                    nested_derivatives += dsp.timeshift(atomic_derivatives, nested_shifts * fs, order)
                atomic_shifts = shifts[(atom, )]
                nested_shifts += dsp.timeshift(atomic_shifts, nested_shifts * fs, order)
            shifts[operator] = nested_shifts
            derivatives[operator] = nested_derivatives
        if delay_derivatives:
            return shifts, derivatives
        return shifts, None

    def build(self, delays, fs, delay_derivatives=None, order=5, delta=1e-12, maxiter=10):
        """Prepare the combination for a set of delays.

        This function prepares the actual combination by pre-computing nested delays and
        Doppler factors. It returns a callable object that evaluates the combination for a
        given set of measurements.

        .. code-block:: python

            built = mytdi.build(
                {'d_12': t12_array, 'd_21': t21_array}, fs,
            )
            tdi_data = built(
                {'y_12': y12_array,'y_21': y21_array},
            )

        If any advancement operators are used in the combination, they are computed
        iteratively.

        Args:
            delays (dict): dictionary of delays [s]
            fs (double): sampling frequency [Hz]
            delay_derivatives: dictionary of delay time derivatives [s/s]
            order (int): Lagrange interpolation orders for delay interpolations
            delta (float): allowed timing accuracy for advancements calculation
            maxiter (int): maximum iterations for advancements calculation

        Returns:
            callable: A callable object that returns the evaluated combinations
            and accepting the following arguments:

            * **measurements** (*dict*): dictionary of measurements
            * **order** (*int*): Lagrange interpolation orders for measurement interpolations
            * **unit** (*str*): unit of measurements, one of ``'phase'``, ``'frequency'``
        """
        logger.info("Building combination '%s'", self)

        shifts, derivatives = self.build_shifts(delays, fs, delay_derivatives, order, delta, maxiter)

        def call(measurements, order=31, unit='frequency'):
            """Evaluate combination for a set of measurements."""
            logger.info("Evaluate combination '%s'", self)

            order = int(order)
            logger.debug("Using interpolation order '%d' for measurements", order)

            result = 0
            for measurement, polynomial in self.components.items():
                logger.debug("Computing contributions from measurement '%s'", measurement)
                for factor, operator in polynomial:
                    # Shift the data, and add it to the result
                    shift = shifts[tuple(operator)]
                    shifted = dsp.timeshift(measurements[measurement], shift * fs, order)
                    if unit == 'phase':
                        result += factor * shifted
                    elif unit == 'frequency':
                        if derivatives:
                            derivative = derivatives[tuple(operator)]
                        else:
                            derivative = dsp.diff(shift, 1 / fs)
                        result += factor * (1 + derivative) * shifted
                    else:
                        raise ValueError(f"Unit '{unit}' not available, choose one of "
                                          "'phase', 'frequency'.")
            return result

        return call

    def transformed(self, mapping):
        """Return a new combination by applying a transformation.

        A transformation is defined by a mapping of indices, e.g. ``{1: 2, 2: 1}``.

        Args:
            mapping (dict): mapping of indices

        Returns:
            A new combination, result of the transformation.
        """
        def transform_string(string, mapping):
            name, i, j = naming.split_indices(string)
            if i not in mapping:
                raise ValueError(f"Incomplete mapping '{mapping}', should contain index '{i}'.")
            if j not in mapping:
                raise ValueError(f"Incomplete mapping '{mapping}', should contain index '{j}'.")
            i, j = mapping[i], mapping[j]
            return naming.join_indices(name, i, j)

        logger.info("Transforming combination '%s' with mapping '%s'", self, mapping)
        transformed = copy.deepcopy(self)
        transformed.components = {}
        for measurement, polynomial in self.components.items():
            transformed_meas = transform_string(measurement, mapping)
            logger.debug("Transforming measurement '%s' into '%s'", measurement, transformed_meas)
            if transformed_meas in transformed.components:
                raise ValueError(f"Invalid mapping '{mapping}', should be a bijection.")
            transformed.components[transformed_meas] = []
            for factor, operators in polynomial:
                transformed_term = (factor, [transform_string(operator, mapping) for operator in operators])
                logger.debug("Transforming term '%s' into '%s'", operators, transformed_term)
                transformed.components[transformed_meas].append(transformed_term)
        return transformed

    def simplified(self):
        """Return a simplified unique representation of the combination.

        We try to apply the following rules, for each time-shift operator polynomial:

            * drop polynomial if vanishing,
            * drop terms with a vanishing factor,
            * collect terms with identifical time-shift operators with a unique factor,
            * drop delay and advancement operator which cancel out.

        We then sort the components of the simplified combination to get a unique representative.

        Returns:
            A new combination, simplified version of the one passed as argument.
        """
        logger.info("Simplifying combination '%s'", self)
        simplified = copy.deepcopy(self)
        for measurement, polynomial in simplified.components.items():
            # Drop variables that are not used
            if not polynomial:
                logger.debug("Empty polynomial for measurement '%s', removing", measurement)
                del simplified.components[measurement]
                return simplified.simplified()
            # Drop term with vanishing factor
            for factor, operators in list(polynomial):
                if not factor:
                    logger.debug("Vanishing factor for term '%s', removing", operators)
                    simplified.components[measurement].remove((factor, operators))
                    return simplified.simplified()
            # Collect terms with identical time-shift operator
            for (factor1, operators1), (factor2, operators2) in itertools.combinations(polynomial, 2):
                if operators1 == operators2:
                    logger.debug("Collecting factors '%s' and '%s' for term '%s'", factor1, factor2, operators1)
                    simplified.components[measurement].remove((factor1, operators1))
                    simplified.components[measurement].remove((factor2, operators2))
                    simplified.components[measurement].append((factor1 + factor2, operators1))
                    return simplified.simplified()
            # Simplify successive delay and advancement operation
            for factor, operators in list(polynomial):
                for i, (operator1, operator2) in enumerate(zip(operators, operators[1:])):
                    name1, i1, j1 = naming.split_indices(operator1)
                    name2, i2, j2 = naming.split_indices(operator2)
                    if (name1, name2) in [("A", "D"), ("D", "A")] and i1 == j2 and j1 == i2:
                        logger.debug("Cancelling operators '%s' and '%s', removing", operator1, operator2)
                        del operators[i + 1]
                        del operators[i]
                        return simplified.simplified()
        # We sort components to return a unique representation
        for measurement in simplified.components:
            simplified.components[measurement].sort()
        return simplified

    def __eq__(self, other):
        simplified_self = self.simplified()
        simplified_other = other.simplified()
        return simplified_self.components == simplified_other.components

    def __ne__(self, other):
        return not self.__eq__(other)

    def __add__(self, other):
        if not isinstance(other, TDICombination):
            TypeError(f"Unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'.")

        logger.info("Adding combinations '%s' and '%s'", self, other)
        summation = copy.deepcopy(self)
        for other_measurement, other_polynomial in other.components.items():
            polynomial = summation.components.get(other_measurement, [])
            summation.components[other_measurement] = polynomial + other_polynomial
        return summation.simplified()

    def __sub__(self, other):
        return self + (-other)

    def __rmul__(self, other):
        if isinstance(other, tuple):
            factor_other, operators_other = other
        elif isinstance(other, list):
            factor_other, operators_other = 1, other
        elif isinstance(other, (int, float)):
            factor_other, operators_other = other, []
        else:
            TypeError(f"unsupported operand type(s) for *: '{type(other)}' and '{type(self)}'.")

        logger.info("Multiplying term '%s %s' and '%s'", factor_other, operators_other, self)
        product = copy.deepcopy(self)
        for measurement, polynomial in self.components.items():
            for factor, operators in polynomial:
                product.components[measurement].remove((factor, operators))
                product.components[measurement].append((factor_other * factor, operators_other + operators))

        return product.simplified()

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return (1 / other) * self
        raise TypeError(f"unsupported operand type(s) for /: '{type(self)}' and '{type(other)}'.")

    def __neg__(self):
        return (-1) * self

    def __matmul__(self, other):
        logger.info("Composing combinations '%s' and '%s'", self, other)
        composition = copy.deepcopy(self)
        for measurement, polynomial in self.components.items():
            if measurement in other:
                del composition.components[measurement]
                for term in polynomial:
                    composition += term * other[measurement]
        return composition.simplified()


class LISATDICombination(TDICombination):
    """Defines a combination for a LISA-like system.

    A LISA-like system possesses the symmetry of the order-6 dihedral group (D3).
    This subclass implements the relevant transformations.

    Args:
        allow_reflections (bool): whether reflection transformations are allowed
        components (dict): dictionary of components, c.f. :class:`pytdi.TDICombination`
    """

    def __init__(self, components, allow_reflections=True):
        super().__init__(components)
        self.allow_reflections = allow_reflections

    def rotated(self, nrot=1):
        """Returns a rotated combination.

        A rotation is a circular permutation of indices.

        Args:
            nrot (int): number of 120-degree rotations

        Returns:
            A new rotated combination.
        """
        nrot = nrot % 3
        logger.info("Rotating combination '%s' by %d degrees", self, 120 * nrot)
        if nrot == 0:
            mapping = {1: 1, 2: 2, 3: 3}
        elif nrot == 1:
            mapping = {1: 2, 2: 3, 3: 1}
        elif nrot == 2:
            mapping = {1: 3, 2: 1, 3: 2}

        return self.transformed(mapping)

    def reflected(self, axis=1):
        """Returns a reflected combination.

        A reflection leaves the axis unchanged, and swaps the others.

        Args:
            axis (int): index around which combination is reflected

        Returns:
            A new reflected combination.
        """
        logger.info("Reflecting combination '%s' around '%d'", self, axis)
        if not self.allow_reflections:
            raise UserWarning("Reflections are not allowed for this combination.")

        if axis == 1:
            mapping = {1: 1, 2: 3, 3: 2}
        elif axis == 2:
            mapping = {1: 3, 2: 2, 3: 1}
        elif axis == 3:
            mapping = {1: 2, 2: 1, 3: 3}
        else:
            raise ValueError(f"Invalid reflection axis '{axis}'.")

        return self.transformed(mapping)


class LISAClockCorrection:
    """Defines a quantity in which clock noise enters similarily to a LISA TDI combination.

    Clock-noise corrections are automatically built from a :class:`pytdi.LISATDICombination`.
    To compute a clock noise-free combination, evaluate the difference of this combination
    and the correction quantity built with this class.


    Args:
        combination (:class:`pytdi.LISATDICombination`): uncorrected combination
        modulation_freqs (dict): constant modulation frequencies [Hz]
        modulation_reduction (bool): removes right-sided modulation noise using RFI sidebands
    """
    # pylint: disable=invalid-name

    def __init__(self, combination, modulation_freqs=None, modulation_reduction=True):

        # Check that `combination` is a LISATDICombination
        if isinstance(combination, LISATDICombination):
            #: :class:`pytdi.LISATDICombination`: Uncorrected combination.
            self.combination = combination
        else:
            raise TypeError(f"invalid combination type {type(combination)}, use LISATDICombination")

        logger.info("Initializing clock-noise correction from combination '%s'", self.combination)

        # Use default set of modulation frequencies, matching that of LISANode and Instrument
        if modulation_freqs is None:
            modulation_freqs = {
                '12': 2.4E9, '23': 2.4E9, '31': 2.4E9,
                '13': 2.401E9, '32': 2.401E9, '21': 2.401E9,
            }

        # Check that only a single measurement name (`var_name`) is found in `combination`
        var_name = None
        for measurement in combination.measurements:
            new_name, _, _ = naming.split_indices(measurement)
            if var_name is None:
                var_name = new_name
            elif var_name != new_name:
                raise ValueError(f"single measurement name allowed, found '{new_name}' and '{var_name}'")

        # Build `Delta_M_ij`
        if modulation_reduction:
            logger.debug("Building the differential modulation noise (ΔM) combinations")
            self.delta_M = {
                i: LISATDICombination({
                    f'rfi_sb_{i}{k}': [(1/2, [])],
                    f'rfi_{i}{k}': [(-1/2, [])],
                    f'rfi_sb_{i}{j}': [(-1/2, [])],
                    f'rfi_{i}{j}': [(1/2, [])],
                })
                for i, j, k in ['123', '231', '312']
            }
        else:
            self.delta_M = {i: LISATDICombination({}) for i in '123'}
        delta_M_set = {f'Delta_M_{i}': self.delta_M[i] for i in self.delta_M}

        # Build `r_ij`
        logger.debug("Building the modulation-corrected clock-correction variables (rc)")
        self.rc = {}
        for i, j, k in ['123', '231', '312']:
            self.rc[f'{i}{j}'] = LISATDICombination({
                f'isi_sb_{i}{j}': [(1 / modulation_freqs[f'{j}{i}'], [])],
                f'isi_{i}{j}': [(-1 / modulation_freqs[f'{j}{i}'], [])],
                f'Delta_M_{j}': [(1 / modulation_freqs[f'{j}{i}'], [f'D_{i}{j}'])],
            }) @ delta_M_set
            self.rc[f'{i}{k}'] = LISATDICombination({
                f'isi_sb_{i}{k}': [(1 / modulation_freqs[f'{k}{i}'], [])],
                f'isi_{i}{k}': [(-1 / modulation_freqs[f'{k}{i}'], [])],
                f'Delta_M_{i}': [(-1 / modulation_freqs[f'{k}{i}'], [])],
            }) @ delta_M_set
        rc_set = {f'rc_{mosa}': rc for mosa, rc in self.rc.items()}

        # Build `P_ij`
        # Note that `P_component` here is denoted `P_ij` in the article, while
        # `P` in this code corresponds to `P_ij rc_ij` in the article
        logger.debug("Building the delay polynomials (P)")
        self.P = {}
        self.P_component = {}
        for i, j in naming.lisa_indices():
            self.P_component[f'{i}{j}'] = self.combination.components.get(f'{var_name}_{i}{j}', [])
            self.P[f'{i}{j}'] = LISATDICombination({f'rc_{i}{j}': self.P_component[f'{i}{j}']}) @ rc_set

        # Build `R_ij`
        logger.debug("Building the reconstructed photon path (R)")
        self.R_rc = {}
        self.R = {}
        for i, j in naming.lisa_indices():
            self.R_rc[f'{i}{j}'] = LISATDICombination({})
            for coeff, delays in self.P_component[f'{i}{j}']:
                self.R_rc[f'{i}{j}'] -= coeff * LISATDICombination.from_string(
                    self.to_string(delays), half_way_shift=False, var_name='rc')
            self.R[f'{i}{j}'] = self.R_rc[f'{i}{j}'] @ rc_set

    @staticmethod
    def to_string(delays):
        """Returns a string from a list of delay operators.

        Args:
            delays (list): delay operators
        """
        components = []
        for delay in delays:
            operator, i, j = naming.split_indices(delay)
            if operator == 'D':
                components.append(f'-{i}{j}')
            elif operator == 'A':
                components.append(f'{i}{j}')
            else:
                raise ValueError(f"unsupported time-shift operator '{operator}', use 'A' or 'D'")
        return ' '.join(components)

    def build(self, *args, **kwargs):
        """Prepare the clock-noise correction from delays.

        This function prepares the actual correction by pre-computing nested delays and
        Doppler factors. It returns a callable object that evaluates the correction for a
        given set of measurements.

        .. code-block:: python

            built = mycorrection.build(
                {'d_12': t12_array, 'd_21': t21_array}, fs,
            )
            correction_data = built(
                {'y_12': y12_array,'y_21': y21_array},
            )

        If any advancement operators are used in the combination, they are computed
        iteratively.

        Args:
            delays (dict): dictionary of delays [s]
            fs (double): sampling frequency [Hz]
            delay_derivatives: dictionary of delay time derivatives [s/s]
            order (int): Lagrange interpolation orders for delay interpolations
            delta (float): allowed timing accuracy for advancements calculation
            maxiter (int): maximum iterations for advancements calculation

        Returns:
            callable: A callable object that returns the evaluated correction
            and accepting the following arguments:

            * **measurements** (*dict*): measurements, including sidebands [Hz]
            * **beatnote_freqs** (*dict*): beatnote frequencies [Hz]
            * **order** (*int*): Lagrange interpolation orders for measurement interpolations
            * **individual_rij** (*bool*): compute ``R_ij`` for each term individually
        """
        logger.info("Building clock-noise correction from combination")

        # Build the `P_ij`
        prepared_P = {
            mosa: P.build(*args, **kwargs)
            for mosa, P in self.P.items()
        }

        # Build the `R_ij`
        prepared_R = {
            mosa: R.build(*args, **kwargs)
            for mosa, R in self.R.items()
        }

        def call(measurements, beatnote_freqs, order=31, individual_rij=True):
            """Evaluate clock-noise correction for a set of measurements."""
            result = 0
            if individual_rij:
                # prepare rescaled measurements
                rescaled_measurements = {}
                for i, j, k in ['123', '231', '312']:
                    rescaled_measurements[f'rfi_{i}{j}'] = {key: beatnote_freqs[f'rfi_{i}{j}'] * value \
                        for key, value in measurements.items()}
                    rescaled_measurements[f'isi_{i}{j}'] = {key: beatnote_freqs[f'isi_{i}{j}'] * value \
                        for key, value in measurements.items()}
                    rescaled_measurements[f'isi_{i}{k}'] = {key: beatnote_freqs[f'isi_{i}{k}'] * value \
                        for key, value in measurements.items()}
                # compute correction
                for i, j, k in ['123', '231', '312']:
                    result += prepared_R[f'{i}{j}'](rescaled_measurements[f'rfi_{j}{k}'], order=order) \
                        - prepared_R[f'{i}{j}'](rescaled_measurements[f'isi_{i}{j}'], order=order) \
                        - prepared_R[f'{i}{k}'](rescaled_measurements[f'rfi_{i}{j}'], order=order) \
                        - prepared_R[f'{i}{k}'](rescaled_measurements[f'isi_{i}{k}'], order=order) \
                        + prepared_P[f'{i}{j}'](rescaled_measurements[f'rfi_{j}{k}'], order=order)
            else:
                # compute correction
                for i, j, k in ['123', '231', '312']:
                    result += (beatnote_freqs[f'rfi_{j}{k}'] - beatnote_freqs[f'isi_{i}{j}']) \
                        * prepared_R[f'{i}{j}'](measurements, order=order) \
                        - (beatnote_freqs[f'rfi_{i}{j}'] + beatnote_freqs[f'isi_{i}{k}']) \
                        * prepared_R[f'{i}{k}'](measurements, order=order) \
                        + beatnote_freqs[f'rfi_{j}{k}'] * prepared_P[f'{i}{j}'](measurements, order=order)
            return result

        return call

    def is_valid(self):
        r"""Test whether clock-noise correction is valid.

        We check symbolically that

        .. math::

            P_{ij} q_i = R_{ij} \qs

        Returns:
            bool: Whether the clock-noise correction is valid.
        """
        test_rc = {
            f'{i}{j}': LISATDICombination({f'q_{i}': [(-1, [])], f'q_{j}': [(1, [f'D_{i}{j}'])]})
            for i, j in naming.lisa_indices()
        }
        test_rc_set = {f'rc_{mosa}': test_rc[mosa] for mosa in test_rc}

        # Check that P_ij q_i = R_ij for all i, j
        for i, j in naming.lisa_indices():
            P_q = LISATDICombination({f'q_{i}': self.P_component[f'{i}{j}']}).simplified()
            if self.R_rc[f'{i}{j}'] @ test_rc_set != P_q:
                return False

        return True

    def __repr__(self):
        return f'<{self.__class__.__name__} based on {self.combination}>'



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Collections of methods for digital signal processing.
"""

import logging
import numpy as np

logger = logging.getLogger(__name__)


def diff(data, dt):
    """Calculate derivative in time.

    Args:
        data (float or array-like): input array
        dt (float): sampling time

    Returns:
        array-like: The array of derived data.
    """
    return 0 if np.isscalar(data) else np.gradient(data, dt)

def timeshift(data, shifts, order=31):
    """Time-shift data (delay or advancement) using Lagrange interpolation.

    The method computes the Lagrange coefficients for the possibly time-varying shifts,
    and apply them as a fractional-delay finite-impulse response Lagrange filter.

    Note:
        The ``shifts`` are given as dimensionless numbers of samples.

    Example:

        .. code-block:: python

            # Apply fractional time-varying delay
            data = np.array(...)
            delays = np.array(...)
            shifted = timeshift(data, -delays, order=11)

            # Apply fractional time-varying advancement
            advancements = np.array(...)
            shifted = timeshift(data, advancements, order=11)

    Args:
        data (float or array-like): input array
        shifts (float or array-like): time shifts [number of samples]
        order (odd int): interpolation order

    Returns:
        array-like: The array of shifted data, whose dimension is adapted to accomodate
        the time shifts.
    """
    if order % 2 == 0:
        raise ValueError(f"`order` must be an odd integer (got {order})")

    data = np.asarray(data)
    shifts = np.asarray(shifts)
    logger.debug("Time shifting data samples (order=%d)", order)

    # Handle constant input and vanishing shifts
    if data.size == 1:
        logger.debug("Input data is constant, skipping time-shift operation")
        return data.item()
    if np.all(shifts == 0):
        logger.debug("Time shifts are vanishing, skipping time-shift operation")
        return data

    logger.debug("Computing Lagrange coefficients")
    halfp = (order + 1) // 2
    shift_ints = np.floor(shifts).astype(int)
    shift_fracs = shifts - shift_ints
    taps = lagrange_taps(shift_fracs, halfp)

    # Handle constant shifts
    if shifts.size == 1:
        logger.debug("Constant shifts, using correlation method")
        i_min = shift_ints - (halfp - 1)
        i_max = shift_ints + halfp + data.size
        if i_max - 1 < 0:
            return np.repeat(data[0], data.size)
        if i_min > data.size - 1:
            return np.repeat(data[-1], data.size)
        logger.debug("Padding data (left=%d, right=%d)", max(0, -i_min), max(0, i_max - data.size))
        data_trimmed = data[max(0, i_min):min(data.size, i_max)]
        data_padded = np.pad(data_trimmed, (max(0, -i_min), max(0, i_max - data.size)), mode='edge')
        logger.debug("Computing correlation product")
        return np.correlate(data_padded, taps[0], mode='valid')

    # Check that sizes or compatible
    if data.size != shifts.size:
        raise ValueError(f"`data` and `shift` must be of the same size (got {data.size}, {shifts.size})")

    # Handle time-varying shifts
    logger.debug("Time-varying shifts, using sliding window view")
    indices = np.clip(np.arange(data.size) + shift_ints, -(halfp + 1), data.size + (halfp - 1))
    padded = np.pad(data, 2 * halfp)
    slices = np.lib.stride_tricks.sliding_window_view(padded, 2 * halfp)
    slices = slices[indices + 2 * halfp - (halfp - 1)]
    logger.debug("Computing matrix-vector product")
    return np.einsum('ij,ij->i', taps, slices)


def lagrange_taps(shift_fracs, halfp):
    """Compute the Lagrange coefficients.

    Computes the Lagrange interpolating polynomial coefficients for a centered
    fractional delay filter.

    Args:
        shift_fracs (array-like): fractional time shifts [samples]
        halfp (int): number of points on each side, equivalent to ``(order + 1) // 2``

    Returns:
        array-like: Array of Lagrange coefficients, of dimension ``(N, 2 * halfp)``
        if N is the dimension of ``shift_fracs``.
    """
    taps = np.zeros((2 * halfp, shift_fracs.size))

    if halfp > 1:
        factor = np.ones(shift_fracs.size, dtype=np.float64)
        factor *= shift_fracs * (1 - shift_fracs)

        for j in range(1, halfp):
            factor *= (-1) * (1 - j / halfp) / (1 + j / halfp)
            taps[halfp - 1 - j] = factor / (j + shift_fracs)
            taps[halfp + j] = factor / (j + 1 - shift_fracs)

        taps[halfp - 1] = 1 - shift_fracs
        taps[halfp] = shift_fracs

        for j in range(2, halfp):
            taps *= 1 - (shift_fracs / j)**2

        taps *= (1 + shift_fracs) * (1 - shift_fracs / halfp)
    else:
        taps[halfp - 1] = 1 - shift_fracs
        taps[halfp] = shift_fracs

    return taps.T

def calculate_advancements(delays, fs, order=5, delta=1e-12, maxiter=10):
    r"""Solve for advancements using delays as inputs.

    This solves iteratively for the advancement :math:`a_{ij}(t)`.

    .. math::

        a_{ij}(t) = d_{ji}(t + a_{ij}(t)) \qc

    with :math:`d_{ji}(t)` being the associated delay.

    Args:
        delays (array-like): time delays [s]
        fs (float): sampling frequency [Hz]
        order (odd int): interpolation order
        delta (float): allowed root mean-square error between two successive iterations
        max_iter (int): maximum number of iterations

    Returns:
        array-like: Array of time advancements, same dimension as ``delays``.
    """
    logger.debug("Computing advancements from delays, solving for implicit equation "
                 "(order=%d, delta=%f, maxiter=%d)", order, delta, maxiter)

    delays = np.asarray(delays)
    if delays.size == 1:
        logger.debug("Delays are constant, skipping advancement calculation")
        return delays

    advancements = delays.copy()
    delta2 = delta**2
    mean_squared_error = delta2

    count = 0
    while mean_squared_error >= delta2:
        logger.debug("Computing iteration %d", count + 1)
        if count >= maxiter:
            logger.warning("Maximum number of iterations %d reached, "
                           "returning current advancement", maxiter)
            break
        advancements_prev = advancements
        advancements = timeshift(delays, advancements_prev * fs, order=order)
        advancement_max = np.max(advancements_prev)
        num_tabs = order + 1
        k_max = int(advancement_max * fs)
        cut_start = max(-(k_max + 1) + num_tabs // 2, 0)
        cut_end = k_max + num_tabs // 2
        difference = (advancements - advancements_prev)[cut_start:-cut_end]
        mean_squared_error = np.mean(difference**2)
        logger.debug("Mean squared error of iteration %d: '%f' (aim at %f)",
            count + 1, mean_squared_error, delta2)
        count += 1

    return advancements



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=missing-module-docstring

__version__ = '1.2.1'
__author__ = 'Martin Staab, Jean-Baptiste Bayle'
__email__ = 'martin.staab@aei.mpg.de, j2b.bayle@gmail.com'
__copyright__ = \
    '2021, Max Planck Institute for Gravitational Physics ' \
    '(Albert Einstein Institute) and California Institute of Technology'



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines TDI first-and-second-generation Michelson variables.

Authors:
    Martin Staab <martin.staab@aei.mpg.de>
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

from . import core
from . import intervar


#: First-generation :math:`X_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
X1_ETA = core.LISATDICombination.from_string('13121 -13121')

#: First-generation :math:`Y_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Y1_ETA = X1_ETA.rotated()


#: First-generation :math:`Z_1` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Z1_ETA = Y1_ETA.rotated()

#: First-generation :math:`X_1` Michelson combination.
#:
#: This combination is the composition of ``X1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
X1 = X1_ETA @ intervar.ETA_SET

#: First-generation :math:`Y_1` Michelson combination.
#:
#: This combination is the composition of ``Y1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Y1 = Y1_ETA @ intervar.ETA_SET

#: First-generation :math:`Z_1` Michelson combination.
#:
#: This combination is the composition of ``Z1_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Z1 = Z1_ETA @ intervar.ETA_SET


#: Second-generation :math:`X_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
X2_ETA = core.LISATDICombination.from_string('131212131 -121313121')

#: Second-generation :math:`Y_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Y2_ETA = X2_ETA.rotated()

#: Second-generation :math:`Z_2` Michelson combination.
#:
#: This combination is defined as a function of the :math:`\eta` intermediary
#: variables. Use them with the inter-spacecraft beatnotes if you want to
#: bypass intermediary variables.
Z2_ETA = Y2_ETA.rotated()

#: Second-generation :math:`X_2` Michelson combination.
#:
#: This combination is the composition of ``X2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
X2 = X2_ETA @ intervar.ETA_SET

#: Second-generation :math:`Y_2` Michelson combination.
#:
#: This combination is the composition of ``Y2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Y2 = Y2_ETA @ intervar.ETA_SET

#: Second-generation :math:`Z_2` Michelson combination.
#:
#: This combination is the composition of ``Z2_ETA`` and the intermediary
#: variables. Therefore, it is function of the beatnote measurements.
Z2 = Z2_ETA @ intervar.ETA_SET



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# BSD 3-Clause License
#
# Copyright (c) 2022, California Institute of Technology and
# Max Planck Institute for Gravitational Physics (Albert Einstein Institute)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# This software may be subject to U.S. export control laws. By accepting this
# software, the user agrees to comply with all applicable U.S. export laws and
# regulations. User has the responsibility to obtain export licenses, or other
# export authority as may be required before exporting such information to
# foreign countries or providing access to foreign persons.
#
"""
Defines convenienve functions to interface with other softwares.

Authors:
    Jean-Baptiste Bayle <j2b.bayle@gmail.com>
"""

import logging
import numpy
import h5py

from scipy.interpolate import InterpolatedUnivariateSpline
from packaging.version import Version
from packaging.specifiers import SpecifierSet

logger = logging.getLogger(__name__)


def _slice(data, skipped):
    """Slice data if it is an array with more than one element.

    * If ``data`` is a scalar, return the scalar.
    * If ``data`` is is one-element array, extract and return it.
    * If ``data`` is any other array, slice it according to `skipped` and return it.

    Args:
        data (scalar or array-like): input array
        skipped (bool): number of samples to skip the beginning

    Returns:
        (scalar or array-like) Sliced data
    """
    if numpy.isscalar(data):
        return data
    if data.size == 1:
        return data[0]
    return data[skipped:]


class Data():
    """Interface object that helps loading data and using them to evaluate combinations.

    The data can be loaded from various sources, including HDF5 files and objects, e.g.,

    .. code-block:: python

        data = Data.from_orbits('my-orbits.h5')
        data = Data.from_lisanode('my-file.h5')
        data = Data.from_instrument(i)

    The data object can then be used to build combinations, either explicitely or using the
    double-star syntax,

    .. code-block:: python

        my_combination.build(data.delays, data.fs)(data.measurements)
        my_combination.build(**data.args)(data.measurements)

    Args:
        measurements (dict): beatnote measurements
        delays (dict): delays [s]
        fs (float): sampling frequency [Hz]
        delay_derivatives (dict): delay time derivatives [s/s]
    """

    #: List of MOSA double indices.
    MOSAS = ['12', '23', '31', '13', '32', '21']

    #: list: List of valid measurements.
    MEASUREMENTS = [
        *[f'isi_{mosa}' for mosa in MOSAS],
        *[f'tmi_{mosa}' for mosa in MOSAS],
        *[f'rfi_{mosa}' for mosa in MOSAS],
        *[f'isi_sb_{mosa}' for mosa in MOSAS],
        *[f'rfi_sb_{mosa}' for mosa in MOSAS],
    ]

    LISANODE_FLUCTUATIONS = {
        **{f'isi_{mosa}': f'isc_c_fluctuations_{mosa}' for mosa in MOSAS},
        **{f'tmi_{mosa}': f'tm_c_fluctuations_{mosa}' for mosa in MOSAS},
        **{f'rfi_{mosa}': f'ref_c_fluctuations_{mosa}' for mosa in MOSAS},
        **{f'isi_sb_{mosa}': f'isc_sb_fluctuations_{mosa}' for mosa in MOSAS},
        **{f'rfi_sb_{mosa}': f'ref_sb_fluctuations_{mosa}' for mosa in MOSAS},
    }

    LISANODE_OFFSETS = {
        **{f'isi_{mosa}': f'isc_c_frequency_{mosa}' for mosa in MOSAS},
        **{f'tmi_{mosa}': f'tm_c_frequency_{mosa}' for mosa in MOSAS},
        **{f'rfi_{mosa}': f'ref_c_frequency_{mosa}' for mosa in MOSAS},
        **{f'isi_sb_{mosa}': f'isc_sb_frequency_{mosa}' for mosa in MOSAS},
        **{f'rfi_sb_{mosa}': f'ref_sb_frequency_{mosa}' for mosa in MOSAS},
    }

    LISANODE_TOTALFREQS = {
        **{f'isi_{mosa}': f'isc_c_{mosa}' for mosa in MOSAS},
        **{f'tmi_{mosa}': f'tm_c_{mosa}' for mosa in MOSAS},
        **{f'rfi_{mosa}': f'ref_c_{mosa}' for mosa in MOSAS},
        **{f'isi_sb_{mosa}': f'isc_sb_{mosa}' for mosa in MOSAS},
        **{f'rfi_sb_{mosa}': f'ref_sb_{mosa}' for mosa in MOSAS},
    }

    LISANODE_MPRS = {f'd_{mosa}': f'mpr_{mosa}' for mosa in MOSAS}

    def __init__(self, measurements, delays, fs, delay_derivatives=None):
        logger.info("Initializing data object")

        self.measurements = measurements
        self.delays = delays
        self.fs = fs
        self.delay_derivatives = delay_derivatives

    @property
    def args(self):
        """Return an object to be used as argument of :meth:`pytdi.TDICombination.build`.

        This includes the delay derivatives if available.

        Returns:
            dict: Dictionary of delays, sampling frequency, and delay derivatives.
        """
        return {
            'delays': self.delays,
            'fs': self.fs,
            'delay_derivatives': self.delay_derivatives,
        }

    def compute_delay_derivatives(self):
        """Compute delay derivatives from delays.

        We use a simple two-point numerical derivative, using :meth:`numpy.gradient`.
        """
        self.delay_derivatives = {}
        logger.info("Computing delay derivatives")
        for key, delay in self.delays.items():
            self.delay_derivatives[key] = numpy.gradient(delay, 1 / self.fs)

    @classmethod
    def from_orbits(cls, path, fs, t0='orbits', dataset='tps/ppr', **kwargs):
        """Load delays from an `orbit file <https://gitlab.in2p3.fr/lisa-simulation/orbits>`_
        and allows to set custom measurement data.

        **Delays** are taken as the proper pseudoranges (PPRs) in the receiver
        spacecraft proper times (TPS) if ``dataset='tps/ppr'``. If ``dataset='tcb/ltt'``, they
        are taken as the light travel times (LTTs) in the barycentric time frame (TCB).

        **Delay derivatives** are computed using :meth:`pytdi.Data.compute_delay_derivatives`.

        At least one **measurement** must be specified as a Numpy array using keyword arguments.
        Valid keywords are specified in :attr:`pytdi.Data.MEASUREMENTS` and include carrier
        and upper-sideband ISI and RFI beatnotes, as well as carrier TMI beatnotes.

        Non-specified measurements are set to 0.

        Example:
            Use this method as follows.

            .. code-block:: python

                data = Data.from_orbits('orbits.h5', fs=4, isi_12=my_array)

        Args:
            path (str): path to orbit file
            fs (float): measurement sampling frequency [Hz]
            t0 (float): measurement initial time [s], or ``'orbits'`` to match that of orbit file
            dataset (str): orbit dataset to use, either ``'tps/ppr'`` or ``'tcb/ltt'``
            **kwargs: non-vanishing measurements, keys must be in :attr:`pytdi.Data.MEASUREMENTS`

        Raises:
            TypeError: if a keyword argument is not a valid measurement
            ValueError: if no measurement specified

        Returns:
            :class:`pytdi.Data`: A data object.
        """
        # Check that we have at least one measurement
        if not kwargs:
            raise ValueError("from_orbits() requires at least one measurement")
        # Check that keywords are valid measurements
        size = 0
        for key, arg in kwargs.items():
            if isinstance(arg, (int, float)):
                size = max(size, 1)
            else:
                size = max(size, len(arg))
            if key not in cls.MEASUREMENTS:
                raise TypeError(f"from_orbits() has invalid measurement key '{key}'")
        # Check that `dataset` is valid
        if dataset not in ['tps/ppr', 'tcb/ltt']:
            raise ValueError(f"invalid dataset '{dataset}', use 'tps/ppr' or 'tcb/ltt'")
        # Build time vector for interpolation
        if t0 == 'orbits':
            with h5py.File(path, 'r') as orbitf:
                # Check orbit file version
                version = Version(orbitf.attrs['version'])
                logger.debug("Using orbit file version %s", version)
                if version.is_devrelease:
                    logger.warning("You are using an orbit file in a development version")
                if version in SpecifierSet('== 1.*', True):
                    attr = 't0' if dataset == 'tcb/ltt' else 'tau0'
                    t0 = float(orbitf.attrs[attr])
                elif version in SpecifierSet('== 2.*', True):
                    t0 = float(orbitf.attrs['t0'])
                else:
                    raise ValueError(f"unsupported lisaorbits version '{version}'")
        logger.debug("Using interpolating time vector (fs=%s, size=%s, t0=%s)", fs, size, t0)
        t = t0 + numpy.arange(size) / fs
        # Load orbits
        delays = {}
        logger.info("Loading orbit file '%s'", path)
        with h5py.File(path, 'r') as orbitf:
            # Check orbit file version
            version = Version(orbitf.attrs['version'])
            logger.debug("Using orbit file version %s", version)
            if version in SpecifierSet('== 1.*', True):
                times = orbitf['tcb']['t'][:] if dataset == 'tcb/ltt' else orbitf['tps']['tau'][:]
                for mosa in cls.MOSAS:
                    values = orbitf[f'tcb/l_{mosa}']['tt'] if dataset == 'tcb/ltt' else orbitf[f'tps/l_{mosa}']['ppr']
                    delays[f'd_{mosa}'] = InterpolatedUnivariateSpline(times, values, k=5, ext='raise')(t)
            elif version in SpecifierSet('== 2.*', True):
                times = orbitf.attrs['t0'] + numpy.arange(orbitf.attrs['size']) * orbitf.attrs['dt']
                values = orbitf[dataset]
                for i, mosa in enumerate(cls.MOSAS):
                    delays[f'd_{mosa}'] = InterpolatedUnivariateSpline(times, values[:, i], k=5, ext='raise')(t)
            else:
                raise ValueError(f"unsupported lisaorbits version '{version}'")
        # Create measurements dictionary
        measurements = {key: kwargs.get(key, 0) for key in cls.MEASUREMENTS}
        # Create instance
        data = cls(measurements, delays, fs)
        data.compute_delay_derivatives()
        return data


    @classmethod
    def from_lisanode(cls, path, signals='fluctuations', skipped=0, central_freq=2.816E14):
        """Load delays and measurement from `LISANode <https://gitlab.in2p3.fr/j2b.bayle/LISANode>`_.

        **Delays** are taken as the measured pseudoranges (MPRs).

        **Delay derivatives** are computed using :meth:`pytdi.Data.compute_delay_derivatives`.

        The **measurements** dictionary is populated with carrier ISI, TMI, and RFI beatnotes,
        as well as the sideband ISI and RFI beatnotes. If ``signals='fluctuations'`` (default behavior),
        the frequency fluctuations are used; if ``signals='offsets'``, the frequency offsets
        are used; if ``signals='total'``, the total frequencies are used.

        Note: LISANode produces fractional frequency fluctuations (unit-less) that are not corrected
        for the beatnote polarity, frequency offsets in MHz and total frequencies in Hz. For compatibility
        reasons the fractional frequency fluctuations are beatnote corrected and converted to frequency
        fluctuations in Hz by multiplying by the sign of the offsets and the central laser frequency.
        Furthermore, the offsets are converted from MHz to Hz by scaling them by 10^6.

        Example:
            Use this method as follows.

            .. code-block:: python

                data = Data.from_lisanode('lisa.h5')

        Args:
            path (str): path to LISANode output file
            signals (str): signal to use, one of ``'fluctuations'``, ``'offsets'``, ``'total'``
            skipped (bool): number of samples to skip the beginning
            central_freq (float): central laser frequency [Hz]

        Returns:
            :class:`pytdi.Data`: A data object, ``measurements`` are in units of Hz.
        """
        delays = {}
        measurements = {}
        logger.info("Opening LISANode output file '%s'", path)
        with h5py.File(path, 'r') as hdf5:
            # Read sampling rate from MPR dataset since it is always present
            any_value = next(iter(cls.LISANODE_MPRS.values()))
            fs = hdf5[any_value].attrs['sampling_frequency'][0]
            # Load measurements
            if signals == 'fluctuations':
                for key in cls.LISANODE_FLUCTUATIONS:
                    key_fluctuations = cls.LISANODE_FLUCTUATIONS[key]
                    key_offsets = cls.LISANODE_OFFSETS[key]
                    # Correct for beatnote polarity and convert to Hz
                    measurements[key] = central_freq * hdf5[key_fluctuations][skipped:, 1] \
                        * numpy.sign(hdf5[key_offsets][skipped:, 1])
            elif signals == 'offsets':
                for key, key_offsets in cls.LISANODE_OFFSETS.items():
                    # Convert from MHz to Hz
                    measurements[key] = 1E6 * hdf5[key_offsets][skipped:, 1]
            elif signals == 'total':
                for key, key_totals in cls.LISANODE_TOTALFREQS.items():
                    measurements[key] = hdf5[key_totals][skipped:, 1]
            else:
                raise ValueError(f"invalid signals parameter '{signals}'")
            # Load MPRs
            for key, key_mprs in cls.LISANODE_MPRS.items():
                delays[key] = hdf5[key_mprs][skipped:, 1]
        logger.debug("Closing LISANode output file '%s'", path)
        # Create instance
        data = cls(measurements, delays, fs)
        data.compute_delay_derivatives()
        return data

    @classmethod
    def from_instrument(cls, instrument_or_path, signals='fluctuations', skipped=0):
        """Load delays and measurement from `LISA Instrument <https://gitlab.in2p3.fr/lisa-simulation/instrument>`_.

        **Delays** are taken as the measured pseudoranges (MPRs).

        **Delay derivatives** are computed using :meth:`pytdi.Data.compute_delay_derivatives`.

        The **measurements** dictionary is populated with carrier ISI, TMI, and RFI beatnotes,
        as well as the sideband ISI and RFI beatnote. If ``signals='fluctuations'`` (default behavior),
        the frequency fluctuations are used; if ``signals='offsets'``, the frequency offsets
        are used; if ``signals='total'``, the total frequencies are used.

        Example:
            Use this method with a :class:`lisainstrument.Instrument` object or a
            measurement file produced with LISA Instrument.

            .. code-block:: python

                instru = Instrument(...)
                instru.write('measuremnets.h5')

                data = Data.from_instrument(instru)
                data = Data.from_instrument('measurements.h5')

        Args:
            instrument_or_path (str or :class:`lisainstrument.Instrument`):
                instrument object or path to measurement file
            signals (str): signal to use, one of ``'fluctuations'``, ``'offsets'``, ``'total'``
            skipped (bool): number of samples to skip the beginning

        Returns:
            :class:`pytdi.Data`: A data object.
        """
        if isinstance(instrument_or_path, str):
            return cls._from_instrument_file(instrument_or_path, signals, skipped)
        if isinstance(instrument_or_path, object):
            return cls._from_instrument_object(instrument_or_path, signals, skipped)
        raise TypeError(f"unsupported object type '{type(instrument_or_path)}'")

    @classmethod
    def _from_instrument_object(cls, instrument, signals='fluctuations', skipped=0):
        """Load data from :class:`lisainstrument.Instrument` object.

        Args:
            instrument (:class:`lisainstrument.Instrument`: instrument object
            signals (str): signal to use, one of ``'fluctuations'``, ``'offsets'``, ``'total'``
            skipped (bool): number of samples to skip the beginning

        Returns:
            :class:`pytdi.Data`: A data object.
        """
        # pylint: disable=too-many-statements
        delays = {}
        measurements = {}
        logger.info("Loading instrument object '%s'", instrument)
        # Check that simulation has been run
        if not instrument.simulated:
            raise ValueError("simulation must be run before loading data")
        # Check version and load data
        version = Version(instrument.version)
        logger.debug("Using lisainstrument version %s", version)
        if version.is_devrelease:
            logger.warning("You are using lisainstrument in a development version")
        if version in SpecifierSet('< 1.1', True):
            fs = instrument.fs
            for mosa in cls.MOSAS:
                # Load measurements
                if signals == 'fluctuations':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isc_carrier_fluctuations[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.ref_carrier_fluctuations[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tm_carrier_fluctuations[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isc_usb_fluctuations[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.ref_usb_fluctuations[mosa], skipped)
                elif signals == 'offsets':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isc_carrier_offsets[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.ref_carrier_offsets[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tm_carrier_offsets[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isc_usb_offsets[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.ref_usb_offsets[mosa], skipped)
                elif signals == 'total':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isc_carriers[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.ref_carriers[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tm_carriers[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isc_usbs[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.ref_usbs[mosa], skipped)
                # Load MPRs
                delays[f'd_{mosa}'] = _slice(instrument.mprs[mosa], skipped)
        elif version in SpecifierSet('< 1.2', True):
            fs = instrument.fs
            for mosa in cls.MOSAS:
                # Load measurements
                if signals == 'fluctuations':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isi_carrier_fluctuations[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.rfi_carrier_fluctuations[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tmi_carrier_fluctuations[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isi_usb_fluctuations[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.rfi_usb_fluctuations[mosa], skipped)
                elif signals == 'offsets':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isi_carrier_offsets[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.rfi_carrier_offsets[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tmi_carrier_offsets[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isi_usb_offsets[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.rfi_usb_offsets[mosa], skipped)
                elif signals == 'total':
                    measurements[f'isi_{mosa}'] = _slice(instrument.isi_carriers[mosa], skipped)
                    measurements[f'rfi_{mosa}'] = _slice(instrument.rfi_carriers[mosa], skipped)
                    measurements[f'tmi_{mosa}'] = _slice(instrument.tmi_carriers[mosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = _slice(instrument.isi_usbs[mosa], skipped)
                    measurements[f'rfi_sb_{mosa}'] = _slice(instrument.rfi_usbs[mosa], skipped)
                # Load MPRs
                delays[f'd_{mosa}'] = _slice(instrument.mprs[mosa], skipped)
        else:
            raise ValueError(f"unsupported lisainstrument version '{version}'")
        # Create instance
        data = cls(measurements, delays, fs)
        data.compute_delay_derivatives()
        return data

    @classmethod
    def _from_instrument_file(cls, path, signals='fluctuations', skipped=0):
        """Load data from LISA Instrument measurement file.

        Args:
            path (str): path to measurement file
            signals (str): signal to use, one of ``'fluctuations'``, ``'offsets'``, ``'total'``
            skipped (bool): number of samples to skip the beginning

        Returns:
            :class:`pytdi.Data`: A data object.
        """
        # pylint: disable=too-many-statements
        delays = {}
        measurements = {}
        logger.info("Loading measurement file '%s'", path)
        with h5py.File(path, 'r') as hdf5:
            # Check version and load data
            version = Version(hdf5.attrs['version'])
            logger.debug("Using measurement file version %s", version)
            if version.is_devrelease:
                logger.warning("You are using a measurement file in a development version")
            if version in SpecifierSet('< 1.1', True):
                fs = hdf5.attrs['fs']
                for mosa in cls.MOSAS:
                    # Load measurements
                    if signals == 'fluctuations':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isc_carrier_fluctuations'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['ref_carrier_fluctuations'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tm_carrier_fluctuations'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isc_usb_fluctuations'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['ref_usb_fluctuations'][mosa], skipped)
                    elif signals == 'offsets':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isc_carrier_offsets'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['ref_carrier_offsets'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tm_carrier_offsets'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isc_usb_offsets'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['ref_usb_offsets'][mosa], skipped)
                    elif signals == 'total':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isc_carriers'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['ref_carriers'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tm_carriers'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isc_usbs'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['ref_usbs'][mosa], skipped)
                    # Load MPRs
                    delays[f'd_{mosa}'] = _slice(hdf5['mprs'][mosa], skipped)
            elif version in SpecifierSet('< 1.2', True):
                fs = hdf5.attrs['fs']
                for mosa in cls.MOSAS:
                    # Load measurements
                    if signals == 'fluctuations':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isi_carrier_fluctuations'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['rfi_carrier_fluctuations'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tmi_carrier_fluctuations'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isi_usb_fluctuations'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['rfi_usb_fluctuations'][mosa], skipped)
                    elif signals == 'offsets':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isi_carrier_offsets'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['rfi_carrier_offsets'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tmi_carrier_offsets'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isi_usb_offsets'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['rfi_usb_offsets'][mosa], skipped)
                    elif signals == 'total':
                        measurements[f'isi_{mosa}'] = _slice(hdf5['isi_carriers'][mosa], skipped)
                        measurements[f'rfi_{mosa}'] = _slice(hdf5['rfi_carriers'][mosa], skipped)
                        measurements[f'tmi_{mosa}'] = _slice(hdf5['tmi_carriers'][mosa], skipped)
                        measurements[f'isi_sb_{mosa}'] = _slice(hdf5['isi_usbs'][mosa], skipped)
                        measurements[f'rfi_sb_{mosa}'] = _slice(hdf5['rfi_usbs'][mosa], skipped)
                    # Load MPRs
                    delays[f'd_{mosa}'] = _slice(hdf5['mprs'][mosa], skipped)
            else:
                raise ValueError(f"unsupported measurement file version '{version}'")
        # Create instance
        data = cls(measurements, delays, fs)
        data.compute_delay_derivatives()
        return data

    @classmethod
    def from_gws(cls, path, orbits, skipped=0, gw_dataset='tps', orbit_dataset='tps/ppr'):
        """Load data from `LISA GW Response <https://gitlab.in2p3.fr/lisa-simulation/gw-response>`_.

        **Delays** are read from an orbit file, c.f. :meth:`pytdi.Data.from_orbits`.
        Use the parameter ``orbit_dataset`` to switch between proper pseudoranges and
        light travel times.

        **Delay derivatives** are computed using :meth:`pytdi.Data.compute_delay_derivatives`.

        **Measurements** are set to zero, except for the carrier and sideband ISI beatnotes,
        which are taken as the GW link responses, read from a GW file.

        Example:
            Use this method as follows.

            .. code-block:: python

                data = Data.from_gws('gws.h5', 'my-orbits.h5', skipped=200)

        Args:
            path (str): path to gravitational-wave file
            orbits (str): path to orbit file
            skipped (bool): number of samples to skip the beginning
            gw_dataset (str): GW dataset to use, either ``'tps'`` or ``'tcb'``
            orbit_dataset (str): orbit dataset to use, either ``'tps/ppr'`` or ``'tcb/ltt'``
        """
        # Warn if inconsistent GW and orbit datasets
        if (gw_dataset == 'tps' and orbit_dataset != 'tps/ppr') \
            or (gw_dataset == 'tcb' and orbit_dataset != 'tcb/ltt'):
            logger.warning("Using inconsistent orbits and GW file datasets ('%s' and '%s')",
                orbit_dataset, gw_dataset)

        # Load measurements
        logger.info("Loading gravitational-wave file '%s'", path)
        with h5py.File(path, 'r') as gwf:
            version = Version(gwf.attrs['version'])
            logger.debug("Using GW file version %s", version)
            if version.is_devrelease:
                logger.warning("You are using a GW file in a development version")
            if version in SpecifierSet('< 1.1', True):
                measurements = {}
                for mosa in cls.MOSAS:
                    measurements[f'isi_{mosa}'] = -_slice(gwf[f'l_{mosa}'], skipped)
                    measurements[f'isi_sb_{mosa}'] = -_slice(gwf[f'l_{mosa}'], skipped)
                    measurements[f'tmi_{mosa}'] = 0.0
                    measurements[f'rfi_{mosa}'] = 0.0
                    measurements[f'rfi_sb_{mosa}'] = 0.0
                fs = gwf.attrs['fs']
                t0 = gwf.attrs['t0']
            elif version in SpecifierSet('== 1.1', True):
                if gw_dataset not in gwf:
                    raise KeyError(f"cannot find dataset '{gw_dataset}' in '{path}'")
                measurements = {}
                for mosa in cls.MOSAS:
                    measurements[f'isi_{mosa}'] = -_slice(gwf[f'{gw_dataset}/l_{mosa}'], skipped)
                    measurements[f'isi_sb_{mosa}'] = -_slice(gwf[f'{gw_dataset}/l_{mosa}'], skipped)
                    measurements[f'tmi_{mosa}'] = 0.0
                    measurements[f'rfi_{mosa}'] = 0.0
                    measurements[f'rfi_sb_{mosa}'] = 0.0
                fs = gwf.attrs['fs']
                t0 = gwf.attrs['t0']
            elif version in SpecifierSet('== 2.*', True):
                if gw_dataset not in gwf:
                    raise KeyError(f"cannot find dataset '{gw_dataset}' in '{path}'")
                measurements = {}
                for imosa, mosa in enumerate(cls.MOSAS):
                    measurements[f'isi_{mosa}'] = -_slice(gwf[f'{gw_dataset}/y'][:, imosa], skipped)
                    measurements[f'isi_sb_{mosa}'] = -_slice(gwf[f'{gw_dataset}/y'][:, imosa], skipped)
                    measurements[f'tmi_{mosa}'] = 0.0
                    measurements[f'rfi_{mosa}'] = 0.0
                    measurements[f'rfi_sb_{mosa}'] = 0.0
                fs = gwf.attrs['fs']
                t0 = gwf.attrs['t0']
            else:
                raise ValueError(f"unsupported GW file version '{version}'")
        # Load delays from orbit file
        return cls.from_orbits(orbits, fs, t0, orbit_dataset, **measurements)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

import setuptools


with open('README.md', 'r') as file:
    long_description = file.read()


meta = {}
with open('pytdi/meta.py') as file:
    exec(file.read(), meta)


setuptools.setup(
    name='pytdi',
    version=meta['__version__'],
    author=meta['__author__'],
    author_email=meta['__email__'],
    description="Python implementation of time-delay interferometry algorithms.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.in2p3.fr/LISA/LDPG/wg6_inrep/pytdi',
    packages=setuptools.find_packages(),
    install_requires=[
        'numpy',
        'packaging',
    ],
    tests_require=['pytest'],
    python_requires='>=3.6',
    license='BSD-3-Clause',
)
